#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include "settings.h"
#include "md5.h"
#include "sha1.h"
#include <libpq-fe.h>
#include <libmemcached/memcached.h>
#include <my_global.h>
#include <mysql.h>

// Define your variables 123

typedef enum {
	t_MYSQL, 
	t_PGSQL
} database_type;

typedef enum {
	t_NONE,
	t_GEO, 
	t_LOAD
} algorithm_type;

struct qdns {
	char type;
	char qname[256];
	char qclass[3];
	char qtype[6];
	int id;
	char qip[256];
};

struct database{
	char user[255];
	char password[255];
	char database[255];
	char host[255];
	unsigned int port;
	char usocket[255];
	database_type type;
};

struct memcached{
	char host[255];
	unsigned int port;
	unsigned int ttl;
};

struct dns{
	char zone[255];
	char email[255];
	char ns1[255];
	char ns2[255];
	char ns3[255];
	char ns4[255];
	char ns5[255];
	char ns6[255];
	char ns7[255];
	char ns8[255];
	char ns9[255];
	short int ttl;
	short int a_ttl;
	short int srv_ttl;
};

struct software{
	char version[32];
	char license[255];
        bool debug;
	struct database database;
	struct dns dns;
	struct memcached memcached;
	algorithm_type algorithm;
	unsigned int numeric_version;
};

typedef struct {
	int ( * init )( );
	int ( * add  )( );
	int ( * calc )( );
	union {
		SHA1_Context   sha1;
//		SHA224_Context sha224;
//		SHA256_Context sha256;
//		SHA384_Context sha384;
//		SHA512_Context sha512;
	} context;
    size_t hashlen;
} sha_t;

struct software instance;
char *strrev(char *str);
char *get_ip_str(const struct sockaddr *sa, char *s, size_t maxlen);
char *resolveHostname(char* hostname);
char *md5(char *unhashed_string);
char *sha1(char *unhashed_string);
algorithm_type selectAlgorithm(char* domain, char* release, char* crypted_license);

char *strrev(char *str){
      char *p1, *p2;

      if (! str || ! *str)
            return str;
      for (p1 = str, p2 = str + strlen(str) - 1; p2 > p1; ++p1, --p2){
            *p1 ^= *p2;
            *p2 ^= *p1;
            *p1 ^= *p2;
      }
      return str;
}

char *get_ip_str(const struct sockaddr *sa, char *s, size_t maxlen){
	switch(sa->sa_family) {
		case AF_INET:
		inet_ntop(AF_INET, &(((struct sockaddr_in *)sa)->sin_addr),
			s, maxlen);
		break;

		case AF_INET6:
		inet_ntop(AF_INET6, &(((struct sockaddr_in6 *)sa)->sin6_addr),
			s, maxlen);
		break;

		default:
		strncpy(s, "Unknown AF", maxlen);
		return NULL;
	}

	return s;
}

char *resolveHostname(char* hostname){
	struct addrinfo hints, *res, *p;
	int status;
	char *ipstr;
	
	if (instance.debug) {
		fprintf(stderr, "DEBUG: char *resolveHostname(char* hostname = %s)\n", hostname);
	}

	memset(&hints, 0, sizeof hints);
	ipstr = malloc(sizeof(char) * INET6_ADDRSTRLEN);
	memset(ipstr, 0, sizeof(char) * INET6_ADDRSTRLEN);
	hints.ai_family = AF_UNSPEC; // AF_INET or AF_INET6 to force version
	hints.ai_socktype = SOCK_STREAM;
	
	if ((status = getaddrinfo(hostname, NULL, &hints, &res)) != 0) {
		if (instance.debug) {
			fprintf(stderr, "DEBUG: getaddrinfo: %s\n", gai_strerror(status));
		}
	}
	else {	  
		for(p = res;p != NULL; p = p->ai_next) {
			//void *addr;
		      	char *ipver;

		      // get the pointer to the address itself,
		      // different fields in IPv4 and IPv6:
			if (p->ai_family == AF_INET) { // IPv4
				struct sockaddr_in *ipv4 = (struct sockaddr_in *)p->ai_addr;
				//addr = &(ipv4->sin_addr);
				ipver = "IPv4";
				
				// convert the IP to a string and print it:
				ipstr = get_ip_str((struct sockaddr *)ipv4, ipstr, sizeof(char)*INET6_ADDRSTRLEN);
				//inet_ntop(p->ai_family, addr, ipstr, sizeof ipstr);
				if (instance.debug) {
					fprintf(stderr, "DEBUG:  %s  (%s): %s\n", hostname, ipver, ipstr);
				}
			} 
	      /*	else { // IPv6
				struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *)p->ai_addr;
				addr = &(ipv6->sin6_addr);
				ipver = "IPv6";
				
				// convert the IP to a string and print it:
				inet_ntop(p->ai_family, addr, ipstr, sizeof ipstr);
				if (instance.debug) {
					fprintf(stderr, "DEBUG:  %s  (%s): %s\n", hostname, ipver, ipstr);
				}
			} */
		}
		freeaddrinfo(res); // free the linked list
	}
	
	return ipstr;
}

char *md5(char *unhashed_string){
	unsigned char *md5_digest;
	unsigned int len;
	struct MD5Context context;
        char *md5_digest_2;
        short int i;
        
	len = strlen(unhashed_string);
	md5_digest = (unsigned char *) malloc(sizeof(unsigned char) * 32);
	memset(md5_digest, 0, sizeof(unsigned char)*32);
	
	MD5Init(&context);
	MD5Update (&context, unhashed_string, len);
	MD5Final (md5_digest, &context);

	md5_digest_2 = (char *) malloc(sizeof(char) * 32);
	memset(md5_digest_2, 0, sizeof(char)*32);
	
	for (i = 0; i < 16; i++)
		sprintf(&md5_digest_2[i*2],"%02x", md5_digest[i]);

	free(md5_digest);
	return md5_digest_2;
}

char *sha1(char *unhashed_string){
	unsigned char *sha_digest;
        sha_t sha;
        char *sha_digest_2;
        short int i;
        
	sha_digest = (unsigned char *) malloc(sizeof(unsigned char) * SHA1_HASH_SIZE);
	memset(sha_digest, 0, sizeof(unsigned char)*SHA1_HASH_SIZE);

	sha.init    = sha1_initialize;
	sha.add     = sha1_add_bytes;
	sha.calc    = sha1_calculate;
	sha.hashlen = SHA1_HASH_SIZE;
	    
	if ( sha.init( &sha.context ) != SHA_DIGEST_OK ){
		fprintf(stderr, "Can not calculate license, aborting!\n");
		return NULL;
	}
	if ( sha.add( &sha.context, unhashed_string, 32 ) != SHA_DIGEST_OK ){
		fprintf(stderr, "Can not calculate license, aborting!\n");
		return NULL;
	}
	if ( sha.calc( &sha.context, sha_digest ) != SHA_DIGEST_OK ){
		fprintf(stderr, "Can not calculate license, aborting!\n");
		return NULL;
	}

	sha_digest_2 = (char *) malloc(sizeof(char) * SHA1_HASH_SIZE * 2);
	memset(sha_digest_2, 0, sizeof(char)*SHA1_HASH_SIZE*2);
	
	for (i = 0; i < sha.hashlen; i++ )
		sprintf(&sha_digest_2[i*2], "%02x", sha_digest[ i ] );
 
	free(sha_digest);
	return sha_digest_2;
}
/*
 * 1.1.x - GEO release-domain
 * 1.1.x - LOAD release&domain
 * 1.2.x - GEO release++domain
 * 1.2.x - LOAD release!!domain

 * 
 */
algorithm_type selectAlgorithm(char* domain, char* release, char* crypted_license){
	algorithm_type currentLicense;
	unsigned int len;
	char *license_string;
        char *geo_hash;
        
	currentLicense = t_NONE;

	// Geo release-domain
	len = strlen(domain);
	license_string = (char *) malloc(sizeof(char) * (len * 2));
	memset(license_string, 0, sizeof(char) * (len * 2));
	strcat(license_string, release);
	strcat(license_string, (const char *)"++\0");
	strcat(license_string, domain);
	len = strlen(license_string);
	
	if (instance.debug) {
		fprintf(stderr, "DEBUG:Geo License string: %s\n", license_string);
	}	
	
	geo_hash = sha1(md5(license_string));

//	if (instance.debug) {
//		fprintf(stderr, "DEBUG:Geo Crypted License string: %s\n", geo_hash);
//	}	

	if (strncmp(crypted_license, geo_hash, SHA1_HASH_SIZE*2) == 0){
		fprintf(stderr, "GEO License is valid\n");
		currentLicense = t_GEO;
	}
	else{
		// Load release&domain
		char *license_string;
                char *load_hash;
                
		len = strlen(domain);
		license_string = (char *) malloc(sizeof(char) * (len * 2));
		memset(license_string, 0, sizeof(char) * (len *2));
		strcat(license_string, release);
		strcat(license_string, (const char *)"!!\0");
		strcat(license_string, domain);
		len = strlen(license_string);
		
		if (instance.debug) {
			fprintf(stderr, "DEBUG:Load License string: %s\n", license_string);
		}	
		
		load_hash = sha1(md5(license_string));

//		if (instance.debug) {
//			fprintf(stderr, "DEBUG:Load Crypted License string: %s\n", load_hash);
//		}	
		
		if (strncmp(crypted_license, load_hash, SHA1_HASH_SIZE*2) == 0){
			fprintf(stderr, "LOAD License is valid\n");
			currentLicense = t_LOAD;
		}
	}
	
	return currentLicense;
}

int main(){
	memcached_server_st *servers = NULL;
	memcached_st *memc;
	memcached_return rc;
	
	FILE *fSettings;
	Settings *settings;
	int result, i;
	char configuration_file[40];

        char helo[4];
	unsigned short int version;
        char trash;
        struct qdns input;
        void *conn = NULL;
        my_bool reconnect = 1;
        short int connected = 0;

	memset(&instance.license, 0, sizeof(char)*255);
	memset(&instance.database.user, 0, sizeof(char)*255);
	memset(&instance.database.password, 0, sizeof(char)*255);
	memset(&instance.database.host, 0, sizeof(char)*255);
	memset(&instance.database.usocket, 0, sizeof(char)*255);
	memset(&instance.dns.zone, 0, sizeof(char)*255);
	memset(&instance.dns.email, 0, sizeof(char)*255);
	memset(&instance.dns.ns1, 0, sizeof(char)*255);
	memset(&instance.dns.ns2, 0, sizeof(char)*255);
	memset(&instance.dns.ns3, 0, sizeof(char)*255);
	memset(&instance.dns.ns4, 0, sizeof(char)*255);
	memset(&instance.dns.ns5, 0, sizeof(char)*255);
	memset(&instance.dns.ns6, 0, sizeof(char)*255);
	memset(&instance.dns.ns7, 0, sizeof(char)*255);
	memset(&instance.dns.ns8, 0, sizeof(char)*255);
	memset(&instance.dns.ns9, 0, sizeof(char)*255);
	memset(&instance.memcached.host, 0, sizeof(char)*255);
	memset(configuration_file, 0, sizeof(char)*40);

	if (access( "latency.pdns.plugin.conf", F_OK ) != -1 ) {
		// file exists
		strcat(configuration_file, (const char *)"latency.pdns.plugin.conf\0");
	} 
	else if (access( "/etc/latency.pdns.plugin.conf", F_OK ) != -1 ) {
		strcat(configuration_file, (const char *)"/etc/latency.pdns.plugin.conf\0");
	}
	else if (access( "/etc/pdns/latency.pdns.plugin.conf", F_OK ) != -1 ) {
		strcat(configuration_file, (const char *)"/etc/pdns/latency.pdns.plugin.conf\0");
	}
	else if (access( "/usr/local/etc/latency.pdns.plugin.conf", F_OK ) != -1 ) {
		strcat(configuration_file, (const char *)"/usr/local/etc/latency.pdns.plugin.conf\0");
	}

	fSettings = fopen(configuration_file, "r");
	if (fSettings == NULL) {
	    fprintf(stderr, "File pbx.dns.ini not found or access denied!\n");
	    exit(1);
	}

	settings = settings_open(fSettings);
	fclose(fSettings);

	if (settings == NULL) {
		fprintf(stderr, "pbx.dns.ini file contents is not valid!\n");
		exit(1);
	}

	if (settings_set(settings, "Application", "Has Started", "True")){
		char *buffer = NULL;
                buffer = malloc(sizeof(char) * 6);
		memset(buffer, 0, sizeof(char)*6);
                instance.debug = false;
		result = settings_get(settings, "Application", "Debug", buffer, sizeof(char)*6);
		if (result == 0) {
			fprintf(stderr, "Debug setting inside Application section is undefined, setting it to false!\n");
			strcpy(buffer, "false");
		}
		else{
                	if (strncmp(buffer,"true", sizeof(char)*strlen("true")) == 0){
				instance.debug = true;
                        	fprintf(stderr, "Debug is on!\n");
			}
			else{
				instance.debug = false;
                        	fprintf(stderr, "Debug is false!\n");
			}
                }
                
		if (instance.debug) {
			fprintf(stderr, "Memcached Server of this application: %s\n", instance.memcached.host);
		}

		result = settings_get(settings, "Application", "Version", instance.version, sizeof(instance.version));
		if (result == 0) {
			fprintf(stderr, "Version setting inside Application section is undefined!\n");
			exit(1);
		}
		else{
                        unsigned int major, minor, revision;
                        major = minor = revision = 0;
                        sscanf(instance.version, "%d.%d.%d", &major, &minor, &revision);
                        instance.numeric_version = major * 10000 + minor * 100 + revision;
                }
		if (instance.debug) {
			fprintf(stderr, "Version of this application: %s:%hu\n", instance.version, instance.numeric_version);
		}
		
		result = settings_get(settings, "Application", "License", instance.license, sizeof(instance.license));
		if (result == 0) {
			fprintf(stderr, "License setting inside Application section is undefined!\n");
			exit(1);
		}
		if (instance.debug) {
			fprintf(stderr, "License of this application: %s\n", instance.license);
		}
                        
	}
	else {
		fprintf(stderr, "There is not section [Application] in the config file!\n");
		exit(1);
	}

	if (settings_set(settings, "DNS", "Has Started", "True")){
		result = settings_get(settings, "DNS", "Zone", instance.dns.zone, sizeof(instance.dns.zone));
		if (result == 0) {
			fprintf(stderr, "Zone setting inside DNS section is undefined!\n");
			exit(1);
		}
		for(i = 0; instance.dns.zone[i]; i++){
			instance.dns.zone[i] = tolower(instance.dns.zone[i]);
		}

		if (instance.debug) {
			fprintf(stderr, "Zone of this application: %s\n", instance.dns.zone);
		}
/////////////////////////////////////////
		result = settings_get(settings, "DNS", "Email", instance.dns.email, sizeof(instance.dns.email));
		if (result == 0) {
			fprintf(stderr, "Email setting inside DNS section is undefined!\n");
			exit(1);
		}
		for(i = 0; instance.dns.email[i]; i++)
			instance.dns.email[i] = tolower(instance.dns.email[i]);

		if (instance.debug) {
			fprintf(stderr, "Email Zone of this application: %s\n", instance.dns.email);
		}
////////////////////////////////////////////////
		result = settings_get(settings, "DNS", "NS1", instance.dns.ns1, sizeof(instance.dns.ns1));
		if (result == 0) {
			fprintf(stderr, "NS1 setting inside DNS section is undefined!\n");
			exit(1);
		}
		for(i = 0; instance.dns.ns1[i]; i++)
			instance.dns.ns1[i] = tolower(instance.dns.ns1[i]);

		if (instance.debug) {
			fprintf(stderr, "NS1 of this application: %s\n", instance.dns.ns1);
		}
//////////////////////////////////////
		result = settings_get(settings, "DNS", "NS2", instance.dns.ns2, sizeof(instance.dns.ns2));
		if (result == 0) {
			fprintf(stderr, "NS2 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns2[i]; i++)
				instance.dns.ns2[i] = tolower(instance.dns.ns2[i]);

		if (instance.debug) {
			fprintf(stderr, "NS2 of this application: %s\n", instance.dns.ns2);
		}
///////////////////////////////////
		result = settings_get(settings, "DNS", "NS3", instance.dns.ns3, sizeof(instance.dns.ns3));
		if (result == 0) {
			fprintf(stderr, "NS3 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns3[i]; i++)
				instance.dns.ns3[i] = tolower(instance.dns.ns3[i]);
		if (instance.debug) {
			fprintf(stderr, "NS3 of this application: %s\n", instance.dns.ns3);
		}
////////////////////////////
		result = settings_get(settings, "DNS", "NS4", instance.dns.ns4, sizeof(instance.dns.ns4));
		if (result == 0) {
			fprintf(stderr, "NS4 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns4[i]; i++)
				instance.dns.ns4[i] = tolower(instance.dns.ns4[i]);

		if (instance.debug) {
			fprintf(stderr, "NS4 of this application: %s\n", instance.dns.ns4);
		}
/////////////////////////////////
		result = settings_get(settings, "DNS", "NS5", instance.dns.ns5, sizeof(instance.dns.ns5));
		if (result == 0) {
			fprintf(stderr, "NS5 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns5[i]; i++)
				instance.dns.ns5[i] = tolower(instance.dns.ns5[i]);

		if (instance.debug) {
			fprintf(stderr, "NS5 of this application: %s\n", instance.dns.ns5);
		}
//////////////////////////////
		result = settings_get(settings, "DNS", "NS6", instance.dns.ns3, sizeof(instance.dns.ns6));
		if (result == 0) {
			fprintf(stderr, "NS6 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns6[i]; i++)
				instance.dns.ns6[i] = tolower(instance.dns.ns6[i]);

		if (instance.debug) {
			fprintf(stderr, "NS6 of this application: %s\n", instance.dns.ns6);
		}
//////////////////////////
			result = settings_get(settings, "DNS", "NS7", instance.dns.ns7, sizeof(instance.dns.ns7));
		if (result == 0) {
			fprintf(stderr, "NS7 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns7[i]; i++)
				instance.dns.ns7[i] = tolower(instance.dns.ns7[i]);

		if (instance.debug) {
			fprintf(stderr, "NS7 of this application: %s\n", instance.dns.ns7);
		}
////////////////////////
		result = settings_get(settings, "DNS", "NS8", instance.dns.ns3, sizeof(instance.dns.ns8));
		if (result == 0) {
			fprintf(stderr, "NS8 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns8[i]; i++)
				instance.dns.ns8[i] = tolower(instance.dns.ns8[i]);

		if (instance.debug) {
			fprintf(stderr, "NS8 of this application: %s\n", instance.dns.ns8);
		}
//////////////////////////////////
		result = settings_get(settings, "DNS", "NS9", instance.dns.ns3, sizeof(instance.dns.ns9));
		if (result == 0) {
			fprintf(stderr, "NS9 setting inside DNS section is undefined!\n");
		}
		else
			for(i = 0; instance.dns.ns9[i]; i++)
				instance.dns.ns9[i] = tolower(instance.dns.ns9[i]);

		if (instance.debug) {
			fprintf(stderr, "NS9 of this application: %s\n", instance.dns.ns9);
		}
/////////////////////////////////////////
		instance.dns.ttl = settings_get_int(settings, "DNS", "TTL");
		if (instance.dns.ttl == 0) {
			fprintf(stderr, "TTL setting inside DNS section is undefined!, using default 3660\n");
			instance.dns.ttl = 3660;
		}
//////////////////////////
		instance.dns.a_ttl = settings_get_int(settings, "DNS", "A_TTL");
		if (instance.dns.a_ttl == 0) {
			fprintf(stderr, "A_TTL setting inside DNS section is undefined!, using default value from TTL\n");
			instance.dns.a_ttl = instance.dns.ttl;
		}
///////////////////////////
		instance.dns.srv_ttl = settings_get_int(settings, "DNS", "SRV_TTL");
		if (instance.dns.srv_ttl == 0) {
			fprintf(stderr, "A_TTL setting inside DNS section is undefined!, using default value from TTL\n");
			instance.dns.srv_ttl = instance.dns.ttl;
		}
////////////////////////////////////
		if (instance.debug) {
			fprintf(stderr, "Default TTL of this application: %d\n", instance.dns.ttl);
			fprintf(stderr, "A TTL of this application: %d\n", instance.dns.a_ttl);
			fprintf(stderr, "SRV TTL of this application: %d\n", instance.dns.srv_ttl);
		}
	}
	else {
		fprintf(stderr, "There is not section [DNS] in the config file!\n");
		exit(1);
	}

	if (settings_set(settings, "DB", "Has Started", "True")){
		char c_buffer[6];
		result = settings_get(settings, "DB", "Server", instance.database.host, sizeof(instance.database.host));
		if (result == 0) {
			fprintf(stderr, "Server setting inside DB section is undefined!\n");
		}
		if (instance.debug) {
			fprintf(stderr, "DB Server of this application: %s\n", instance.database.host);
		}	
///////////////////////
		result = settings_get(settings, "DB", "Database", instance.database.database, sizeof(instance.database.database));
		if (result == 0) {
			strcpy(instance.database.database, "latency");
			fprintf(stderr, "Database setting inside DB section is undefined! Using default name %s\n", instance.database.database);
		}
		if (instance.debug) {
			fprintf(stderr, "DB Database of this application: %s\n", instance.database.database);
		}	
///////////////////////
		result = settings_get(settings, "DB", "User", instance.database.user, sizeof(instance.database.user));
		if (result == 0) {
			fprintf(stderr, "User setting inside DB section is undefined!\n");
			exit(1);
		}
		if (instance.debug) {
			fprintf(stderr, "DB User of this application: %s\n", instance.database.user);
		}	
//////////////////////////////
		result = settings_get(settings, "DB", "Password", instance.database.password, sizeof(instance.database.password));
		if (result == 0) {
			fprintf(stderr, "Password setting inside DB section is undefined!\n");
		}
		if (instance.debug) {
			fprintf(stderr, "DB Password of this application: %s\n", instance.database.password);
		}	
//////////////////////////
		//int settings_get_int(const Settings *settings, const char *section, const char *key)
		instance.database.port = settings_get_int(settings, "DB", "Port");
		if (instance.database.port == 0) {
			fprintf(stderr, "Port setting inside DB section is undefined!\n");
			instance.database.port = 3306;
		}
		if (instance.debug) {
			fprintf(stderr, "DB Port of this application: %d\n", instance.database.port);
		}	
////////////////////////
		result = settings_get(settings, "DB", "Unix Socket", instance.database.usocket, sizeof(instance.database.usocket));
		if (result == 0) {
			fprintf(stderr, "Unix Socket setting inside DB section is undefined!\n");
		}
		if (instance.debug) {
			fprintf(stderr, "DB Unix Socket of this application: %s\n", instance.database.usocket);
		}
////////////////////////////
		if (((strlen(instance.database.host)) || (strlen(instance.database.usocket))) == 0){
			fprintf(stderr, "You must define Server or Unix Socket setting inside MySQL section. Aborting!\n");
			exit(1);
		}
////////////////////////
		result = settings_get(settings, "DB", "Type", c_buffer, sizeof(c_buffer));
		if (result == 0) {
			fprintf(stderr, "Type setting inside DB section is undefined, put mysql or pgsql!\n");
			exit(1);
		}
		else {
			if (strncmp(c_buffer,"mysql", sizeof(char)*strlen("mysql")) == 0){
				instance.database.type = t_MYSQL;
			}
			else if (strncmp(c_buffer,"pgsql", sizeof(char)*strlen("pgsql")) == 0){
				instance.database.type = t_PGSQL;
			}
			else{
				fprintf(stderr, "Type setting inside DB section is undefined, put mysql or pgsql!\n");
				exit(1);
			}
		}
		if (instance.debug) {
			fprintf(stderr, "DB Type of this application: %s\n", (instance.database.type==t_MYSQL?"MySQL":"PgSQL"));
		}
	}
	else {
		fprintf(stderr, "There is not section [DB] in the config file!\n");
		exit(1);
	}
	
	if (settings_set(settings, "Memcached", "Has Started", "True")){
		result = settings_get(settings, "Memcached", "Server", instance.memcached.host, sizeof(instance.memcached.host));
		if (result == 0) {
			fprintf(stderr, "Server setting inside Memcached section is undefined, setting it to 127.0.0.1!\n");
			strcpy(instance.memcached.host, "127.0.0.1");
		}
		if (instance.debug) {
			fprintf(stderr, "Memcached Server of this application: %s\n", instance.memcached.host);
		}	
////////////////////////
		instance.memcached.port = settings_get_int(settings, "Memcached", "Port");
		if (instance.memcached.port == 0) {
			fprintf(stderr, "Port setting inside Memcached section is undefined!\n");
			instance.memcached.port = 11211;
		}
		if (instance.debug) {
			fprintf(stderr, "Memcached Port of this application: %d\n", instance.memcached.port);
		}
/////////////////////////
		instance.memcached.ttl = settings_get_int(settings, "Memcached", "TTL");
		if (instance.memcached.ttl == 0) {
			fprintf(stderr, "TTL setting inside Memcached section is undefined!\n");
			instance.memcached.ttl = 60;
		}
		if (instance.debug) {
			fprintf(stderr, "Memcached TTL of this application: %d\n", instance.memcached.ttl);
		}	
	}
	else{
		strcpy(instance.memcached.host, "127.0.0.1");
		instance.memcached.port = 11211;
		instance.memcached.ttl = 60;
		if (instance.debug) {
			fprintf(stderr, "Putting Memcached settings by default\n");
		}	
	}

	if ((instance.algorithm = selectAlgorithm(instance.dns.zone, instance.version, instance.license)) != t_NONE){
		fprintf(stderr, "License is valid\n");
	}
	else{
		fprintf(stderr, "License is NOT valid\n");
		fprintf(stderr, "Please send an email to info@okay.com.mx for more information on how to get a valid license\n");
		exit(1);
	}

	memc = memcached_create(NULL);
	servers = memcached_server_list_append(servers, instance.memcached.host, instance.memcached.port, &rc);
	rc = memcached_server_push(memc, servers);
	
	if (instance.debug) {
                if (rc == MEMCACHED_SUCCESS)		
                        fprintf(stderr, "Memcached server successfully added\n");
                else{
                        fprintf(stderr, "I couldn't add memcached server: %s\n", memcached_strerror(memc, rc));
                }
	}
	
	// Until we get HELO\t1
	setbuf(stdout, NULL);
	fflush(stdin);
	do {
		if (scanf("%s\t%hu", helo, &version) != EOF){
                    if (instance.debug) {
                        fprintf(stderr, "DEBUG:[%s]:[%hu]\n", helo, version);
                    }
                }
                else {
                    printf ("LOG\tPowerDNS sent unparseable line\n");
                    printf ("FAIL\n");
                }
	} while (version != 1);
	printf ("OK\tLow latency DNS pluging 1.2 starting now\n");
	
	fflush(stdin);
	if (scanf("%c", &trash) == EOF){
                printf ("LOG\tPowerDNS sent unparseable line step 2\n");
                printf ("FAIL\n");
        }
	fflush(stdin);

	memset(&input.qname, 0, sizeof(char)*256);
	memset(&input.qclass, 0, sizeof(char)*3);
	memset(&input.qtype, 0, sizeof(char)*6);
	memset(&input.qip, 0, sizeof(char)*256);
                
        // Connecting to DB
        if (instance.debug) {
                fprintf(stderr, "DEBUG:Connecting to Database...");
        }

        if (instance.database.type==t_MYSQL){
				if (instance.debug) {
					fprintf(stderr, "MySQL");
				}
                conn = mysql_init(NULL);
                mysql_options(conn, MYSQL_OPT_RECONNECT, &reconnect);
                connected = (mysql_real_connect((MYSQL*)conn, instance.database.host, instance.database.user, instance.database.password, instance.database.database, instance.database.port, NULL, 0)) ||
                            (mysql_real_connect((MYSQL*)conn, NULL, instance.database.user, instance.database.password, instance.database.database, instance.database.port, instance.database.usocket, 0));
        }
        else{
				if (instance.debug) {
					fprintf(stderr, "PostgreSQL");
				}
                char connection_string[255];
                memset(&connection_string, 0, sizeof(char)*255);
                sprintf(connection_string, 
                        "host=%s port=%i user=%s password=%s dbname=%s", 
                        instance.database.host, instance.database.port,
                        instance.database.user, instance.database.password,
                        instance.database.database);
                conn = PQconnectdb(connection_string);
                if (PQstatus((PGconn*)conn) != CONNECTION_BAD){
                        connected = 1;
                }
                else {
                        memset(&connection_string, 0, sizeof(char)*255);
                        sprintf(connection_string, 
                                "host=%s user=%s password=%s dbname=%s", 
                                instance.database.usocket,
                                instance.database.user, instance.database.password,
                                instance.database.database);
                        conn = PQconnectdb(connection_string);
                        if (PQstatus((PGconn*)conn) != CONNECTION_BAD){
                                connected = 1;
                        }
                }
        }

        if (instance.debug) {
                fprintf(stderr, "%i\n", connected);
        }

	while(scanf("%c%s%s%s%d%s", &input.type, input.qname, input.qclass, input.qtype, &input.id, input.qip)) {
		unsigned short int i;

                if (scanf("%c", &trash) == EOF){
                        printf ("LOG\tPowerDNS sent unparseable line step 3\n");
                        printf ("FAIL\n");
                }
		fflush(stdin);

		// All to lowcase
		input.type = tolower(input.type);
		for(i = 0; input.qname[i]; i++){
		    input.qname[i] = tolower(input.qname[i]);
		}
		for(i = 0; input.qclass[i]; i++){
		    input.qclass[i] = tolower(input.qclass[i]);
		}
		for(i = 0; input.qtype[i]; i++){
		    input.qtype[i] = tolower(input.qtype[i]);
		}
		for(i = 0; input.qip[i]; i++){
		    input.qip[i] = tolower(input.qip[i]);
		}
		
		if (instance.debug) {
			fprintf(stderr, "DEBUG:[%c]\t[%s]\t[%s]\t[%s]\t[%d]\t[%s]\n", input.type, input.qname, input.qclass, input.qtype, input.id, input.qip);
		}
		printf ("LOG\t[%c]\t[%s]\t[%s]\t[%s]\t[%d]\t[%s]\n", input.type, input.qname, input.qclass, input.qtype, input.id, input.qip);

		if (!strlen(input.qip)){
			// There is no query to look for in i
			printf ("LOG\tPowerDNS sent unparseable line\n");
			printf ("FAIL\n");
		}
		else if(strstr(input.qname, instance.dns.zone) == NULL){
			printf ("LOG\t%s is not allowed\n", input.qname);
			printf ("FAIL\n");
		}
		else{
			char key[256];
			size_t value_length;
			uint32_t flags;
                        
			memset(key, 0, sizeof(key));
			// key format: latency_<ip>
			strcpy(key, "latency_");
			strcat(key, input.qip);
			if (instance.debug)
				fprintf(stderr, "DEBUG: memcached base key is %s\n", key);
			
			if (strstr(input.qname, instance.dns.zone) != NULL){
				char query[1024];
				memset(query, 0, sizeof(char) * 1024);
				
				if (instance.debug)
					fprintf(stderr, "DEBUG: Looking in the %s zone\n", instance.dns.zone);
	
				if (connected){
					short int executed = 0;
					void *resource = NULL;
					int ipcount = 0;

					if (instance.debug)
						fprintf(stderr, "DEBUG: We are still connected\n");

					// Look if the given IP has a record
					if (instance.algorithm == t_GEO){
						sprintf(query, "SELECT count(*) AS c FROM o_times WHERE o_times.domain='%s'", input.qname);
						if (instance.database.type == t_MYSQL){
							executed = !mysql_query((MYSQL*)conn, query) && (resource = mysql_use_result((MYSQL*)conn));
						}
						else{
							resource = PQexec((PGconn*)conn, query);
							executed = PQresultStatus((PGresult*)resource) == PGRES_TUPLES_OK?1:0;
						}

						if (instance.debug) {
							fprintf(stderr, "DEBUG: %s = executed %hu\n", query, executed);
						}
					}
					else if(instance.algorithm == t_LOAD){
						executed = 1;
						if (instance.debug) {
							fprintf(stderr, "DEBUG: LOAD does not need to review the times table\n");
						}
					}

					if (executed){
						if (instance.algorithm == t_GEO){
							if (instance.database.type==t_MYSQL){
								MYSQL_ROW *row;
								while ((row = (MYSQL_ROW *)mysql_fetch_row((MYSQL_RES*)resource)) != NULL){
									ipcount = atoi((const char*)(row[0]));
									if (instance.debug) {
										fprintf(stderr, "DEBUG: MYSQL result: %s\n",(char*)(row[0]));
									}
								}
								mysql_free_result((MYSQL_RES*)resource);
							}
							else{
								int rows;
								int i;
								rows = PQntuples((PGresult*)resource);
								for (i = 0; i < rows; i++){
									ipcount = atoi((const char*)PQgetvalue((PGresult*)resource,i,0));
									if (instance.debug) {
										fprintf(stderr, "DEBUG: PGSQL result: %s\n",(char*)PQgetvalue((PGresult*)resource,i,0));
									}
								}
								PQclear((PGresult*)resource);
							}
							if (instance.debug) {
								fprintf(stderr, "DEBUG: Total Count = %hu\n", ipcount);
							}
						}
						else if (instance.algorithm == t_LOAD){
							ipcount = 1; // t_LOAD does not need this
							if (instance.debug) {
								fprintf(stderr, "DEBUG: LOAD does not need to count\n");
							}
						}

						if (ipcount){
							if (instance.debug) {
								fprintf(stderr, "DEBUG: Count > 0\n");
							}
							short int found_a = 0;
							short int found_srv = 0;

							if ((strncmp(input.qtype, "any", 3) == 0) || (strncmp(input.qtype, "a", 1) == 0)){
								//char *tokenPtr;
								//tokenPtr = strtok(input.qname, ".");
								char *key_a = NULL;
								char *ipstr;

								key_a = malloc(sizeof(char) * 256);
								memset(key_a, 0, sizeof(char)*256);
								strcpy(key_a, key);
								strcat(key_a, "_a");

								if (instance.debug)
									fprintf(stderr, "DEBUG: Looking for A record\n");

								ipstr = memcached_get(memc, key_a, strlen(key_a), &value_length, &flags, &rc);
								if (rc == MEMCACHED_SUCCESS) {
									found_a++;
									if (instance.debug){
										fprintf(stderr, "DEBUG: no need to set key: %s, already in memcached\n", key);
										fprintf(stderr, "DEBUG: A record found in Memcached\n");
									}
								}
								else{
									resource = NULL;
									executed = 0;

									// key_a has the right value to look force
									if (instance.database.type == t_MYSQL){
										if (instance.algorithm == t_LOAD)
											sprintf(query, "SELECT t.hostname, SUM(t.load_index) AS load_index FROM (SELECT hostname, COUNT(hostname) * 1.5 AS load_index FROM channels INNER JOIN o_servers ON channels.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT hostname, COUNT(hostname) AS load_index FROM registrations INNER JOIN o_servers ON registrations.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT DISTINCT fqdn as hostname, 0 AS load_index FROM o_servers where enabled = 1 AND status = 2) t GROUP BY hostname ORDER BY load_index ASC LIMIT 1");
										else if (instance.algorithm == t_GEO)
											sprintf(query, "SELECT o_servers.fqdn, COUNT(DISTINCT(user)) AS total_extensions, AVG(delay) AS average_delay FROM o_times,o_servers WHERE o_times.domain='%s' AND o_servers.enabled=1 AND o_servers.status=2 AND o_servers.fqdn = o_times.fqdn  GROUP BY fqdn ORDER BY total_extensions DESC, average_delay ASC LIMIT 1", input.qname);
										//	sprintf(query, "SELECT o_servers.fqdn, AVG(delay) AS average_delay FROM o_times,o_servers WHERE o_servers.enabled=1 AND o_servers.status=2 AND o_servers.fqdn = o_times.fqdn AND (INET_ATON('%s') BETWEEN network AND broadcast) GROUP BY fqdn ORDER BY average_delay ASC LIMIT 1", input.qip);
										executed = !mysql_query((MYSQL*)conn, query) && (resource = mysql_use_result((MYSQL*)conn));
									}
									else{
										if (instance.algorithm == t_LOAD)
											sprintf(query, "SELECT t.hostname, SUM(t.load_index) AS load_index FROM (SELECT hostname, COUNT(hostname) * 1.5 AS load_index FROM channels INNER JOIN o_servers ON channels.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT hostname, COUNT(hostname) AS load_index FROM registrations INNER JOIN o_servers ON registrations.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT DISTINCT fqdn as hostname, 0 AS load_index FROM o_servers where enabled = 1 AND status=2) t GROUP BY hostname ORDER BY load_index ASC LIMIT 1");
										else if (instance.algorithm == t_GEO)
											sprintf(query, "SELECT o_servers.fqdn, COUNT(DISTINCT user) AS total_extensions, AVG(delay) AS average_delay FROM o_times,o_servers WHERE o_times.domain='%s' AND o_servers.enabled=1 AND o_servers.status=2 AND o_servers.fqdn = o_times.fqdn  GROUP BY fqdn ORDER BY total_extensions DESC, average_delay ASC LIMIT 1", input.qname);
										//	sprintf(query, "SELECT o_servers.fqdn, AVG(delay) AS average_delay FROM o_times,o_servers WHERE o_servers.enabled=1 AND o_servers.status=2 AND o_servers.fqdn = o_times.fqdn AND (INET '%s' > network AND INET '%s' < broadcast) GROUP BY fqdn ORDER BY average_delay ASC LIMIT 1", input.qip, input.qip);
										
										resource = PQexec((PGconn*)conn, query);
										executed = PQresultStatus((PGresult*)resource) == PGRES_TUPLES_OK?1:0;
									}
									
									if (instance.debug) {
										fprintf(stderr, "DEBUG: %s\n", query);
									}
									
									if (executed){
										char *hostname_a;
										hostname_a = malloc(sizeof(char) * 256);
										memset (hostname_a, 0, sizeof(char) * 256);

										if (instance.database.type==t_MYSQL){
											MYSQL_ROW *row;
											while ((row = (MYSQL_ROW *)mysql_fetch_row((MYSQL_RES*)resource)) != NULL){
												strcpy(hostname_a, (char*)row[0]);
												if (instance.debug) {
													fprintf(stderr, "DEBUG: MYSQL result: %s\n",(char*)row[0]);
												}
												found_a++;
											}
											mysql_free_result((MYSQL_RES*)resource);
										}
										else{
											int rows;
											int i;
											rows = PQntuples((PGresult*)resource);
											for (i = 0; i < rows; i++){
												strcpy(hostname_a, PQgetvalue((PGresult*)resource,i,0));
												if (instance.debug) {
													fprintf(stderr, "DEBUG: PGSQL result: %s\n",(char*)PQgetvalue((PGresult*)resource,i,0));
												}
												found_a++;
											}
											PQclear((PGresult*)resource);
										}
										
										if (found_a){
												if (instance.debug) {
													fprintf(stderr, "DEBUG: Found a record\n");
												}

												if ((ipstr = resolveHostname(hostname_a)) == NULL){
													printf ("LOG\tCouldnt resolve %s\n", hostname_a);
													printf ("FAIL\n");
													found_a = 0;
												}
												else{
													rc = memcached_set(memc, key_a, strlen(key_a), ipstr, strlen(ipstr), (time_t)instance.memcached.ttl, (uint32_t)0);
													if (instance.debug) {
														if (rc == MEMCACHED_SUCCESS) 
															fprintf(stderr, "DEBUG: memcached key: %s set with value '%s'\n", key_a, ipstr);
														else
															fprintf(stderr, "DEBUG: memcached key: %s was not able to be set into memcached with value '%s'\n", key_a, ipstr);
													}
												}
										}
										free(hostname_a);
									}
								}

								if(found_a){
									if (instance.debug) {
										fprintf(stderr, "DEBUG:DATA\t%s\tin\tA\t%i\t1\t%s\n", input.qname, instance.dns.a_ttl, (char *)ipstr);
									}
									printf ("DATA\t%s\tin\tA\t%i\t1\t%s\n", input.qname, instance.dns.a_ttl, (char *)ipstr);
								}
								free(key_a);
								free(ipstr);  // only free ipstr if it doen't come from memcached
							} // if ((strncmp(input.qtype, "any", 3) == 0) || (strncmp(input.qtype, "a", 1) == 0))

							if ((instance.numeric_version >= 10200) &&	
								((strncmp(input.qtype, "any", 3) == 0) || (strncmp(input.qtype, "srv", 3) == 0)) &&
								((strstr(input.qname, "._tcp.") != NULL) || (strstr(input.qname, "._udp.") != NULL) || (strstr(input.qname, "._tls.") != NULL))
								){
								//char *tokenPtr;
								//tokenPtr = strtok(input.qname, ".");
								char *key_srv = NULL;
								char *ipstr;
								short int srv_total = 1;
								char *key_srv_seq = NULL;
								char *key_srv_type = NULL;

								if (instance.debug)
									fprintf(stderr, "DEBUG: Looking for SRV record\n");

								key_srv = malloc(sizeof(char) * 256);
								memset(key_srv, 0, sizeof(char) * 256);

								key_srv_seq = malloc(sizeof(char) * 256);
								memset(key_srv_seq, 0, sizeof(char) * 256);

								key_srv_type = malloc(sizeof(char) * 256);
								memset(key_srv_type, 0, sizeof(char) * 256);

								strcpy(key_srv, key);
								strcat(key_srv, "_srv");

								if (strstr(input.qname, "._udp.") != NULL){
									strcat(key_srv_type, "_udp");
								}
								else if ((strstr(input.qname, "_sips._tcp.") != NULL) || (strstr(input.qname, "._tls.") != NULL)){
									strcat(key_srv_type, "_tls");
								}
								else if (strstr(input.qname, "._tcp.") != NULL){
									strcat(key_srv_type, "_tcp");
								}
								else{
									strcat(key_srv_type, "_*");
								}

								strcat(key_srv, key_srv_type);

								sprintf(key_srv_seq, "%s_%hu", key_srv, srv_total);
								if (instance.debug){
									fprintf(stderr, "DEBUG: memcached SRV base key is %s\n", key_srv);
									fprintf(stderr, "DEBUG: first key %s\n", key_srv_seq);
								}

								for (rc = MEMCACHED_SUCCESS; rc == MEMCACHED_SUCCESS; sprintf(key_srv_seq, "%s_%hu", key_srv, ++srv_total)){
									ipstr = memcached_get(memc, key_srv_seq, strlen(key_srv_seq), &value_length, &flags, &rc);

									if (rc == MEMCACHED_SUCCESS){
										if (instance.debug) {
											fprintf(stderr, "DEBUG:From Memcached\n");
											fprintf(stderr, "DEBUG:DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
											fprintf(stderr, "DEBUG:DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
										}
										printf ("DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
										printf ("DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
									}                                                        
								}

								if (srv_total > 2) {
									found_srv++;
									srv_total-=2;
									if (instance.debug){
										fprintf(stderr, "DEBUG: SRV record found %hu in Memcached\n", srv_total);
										fprintf(stderr, "DEBUG: no need to set key: %s, already in memcached\n", key);
									}
								}
								else{
									void *resource = NULL;
									short int executed = 0;
									srv_total = 0;
									// key_a has the right value to look force
									if (instance.database.type == t_MYSQL){
										if (instance.algorithm == t_LOAD)
											sprintf(query, "SELECT t.hostname, SUM(t.load_index) AS load_index FROM (SELECT hostname, COUNT(hostname) * 1.5 AS load_index FROM channels INNER JOIN o_servers ON channels.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT hostname, COUNT(hostname) AS load_index FROM registrations INNER JOIN o_servers ON registrations.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT DISTINCT fqdn as hostname, 0 AS load_index FROM o_servers where enabled = 1 AND status = 2) t GROUP BY hostname ORDER BY load_index ASC");
										else if (instance.algorithm == t_GEO)
											sprintf(query, "SELECT o_servers.fqdn, (SELECT AVG(delay) AS average_delay FROM o_servers s LEFT OUTER JOIN o_times ON s.fqdn = o_times.fqdn WHERE s.enabled = 1 AND s.status = 2 AND (INET_ATON('%s') BETWEEN network AND broadcast) AND o_servers.group = s.group GROUP BY o_servers.group) as average_delay FROM o_servers WHERE o_servers.enabled=1 AND o_servers.status = 2 ORDER BY average_delay ASC", input.qip);
//											sprintf(query, "SELECT o_servers.fqdn, AVG(delay) AS average_delay FROM o_servers LEFT OUTER JOIN o_times ON o_servers.fqdn = o_times.fqdn WHERE o_servers.enabled=1 AND o_servers.status=2 AND (INET_ATON('%s') BETWEEN network AND broadcast) GROUP BY fqdn UNION SELECT o_servers.fqdn, AVG(delay) AS average_delay FROM o_servers LEFT OUTER JOIN o_times ON o_servers.fqdn = o_times.fqdn WHERE o_servers.enabled=1 AND o_servers.status=2  GROUP BY fqdn ORDER BY average_delay ASC", input.qip);
										executed = !mysql_query((MYSQL*)conn, query) && (resource = mysql_use_result((MYSQL*)conn));
									}
									else{
										if (instance.algorithm == t_LOAD)
											sprintf(query, "SELECT t.hostname, SUM(t.load_index) AS load_index FROM (SELECT hostname, COUNT(hostname) * 1.5 AS load_index FROM channels INNER JOIN o_servers ON channels.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT hostname, COUNT(hostname) AS load_index FROM registrations INNER JOIN o_servers ON registrations.hostname = o_servers.fqdn WHERE o_servers.enabled = 1 AND o_servers.status = 2 GROUP BY hostname UNION SELECT DISTINCT fqdn as hostname, 0 AS load_index FROM o_servers where enabled = 1 AND status = 2) t GROUP BY hostname ORDER BY load_index ASC");
										else if (instance.algorithm == t_GEO)
												sprintf(query, "SELECT o_servers.fqdn, (SELECT AVG(delay) AS average_delay FROM o_servers s LEFT OUTER JOIN o_times ON s.fqdn = o_times.fqdn WHERE s.enabled=1 AND s.status=2 AND (INET '%s' > network AND INET '%s' < broadcast) AND o_servers.group = s.group GROUP BY o_servers.group) as average_delay FROM o_servers WHERE o_servers.enabled=1 AND o_servers.status=2 ORDER BY average_delay ASC", input.qip, input.qip);
//												sprintf(query, "SELECT o_servers.fqdn, AVG(delay) AS average_delay FROM o_servers LEFT OUTER JOIN o_times ON o_servers.fqdn = o_times.fqdn WHERE o_servers.enabled=1 AND o_servers.status=2 AND (INET '%s' > network AND INET '%s' < broadcast) GROUP BY fqdn UNION SELECT o_servers.fqdn, AVG(delay) AS average_delay FROM o_servers LEFT OUTER JOIN o_times ON o_servers.fqdn = o_times.fqdn WHERE o_servers.enabled=1 AND o_servers.status=2  GROUP BY fqdn ORDER BY average_delay ASC", input.qip, input.qip);
										resource = PQexec((PGconn*)conn, query);
										executed = PQresultStatus((PGresult*)resource) == PGRES_TUPLES_OK?1:0;
									}

									if (instance.debug) {
										fprintf(stderr, "DEBUG: %s\n", query);
									}

									if (executed){
										char *hostname_srv;
										int average = 1;
										hostname_srv = malloc(sizeof(char) * 256);
										memset (hostname_srv, 0, sizeof(char) * 256);
										ipstr = malloc(sizeof(char) * 256);
										memset(ipstr, 0, sizeof(char) * 256);

										if (instance.database.type==t_MYSQL){
											unsigned int i;
											MYSQL_ROW *row;

											for (i = 1; (row = (MYSQL_ROW *)mysql_fetch_row((MYSQL_RES*)resource)) != NULL; i++){
												strcpy(hostname_srv, (char*)row[0]);

												if (row[1]){
													average = atoi((const char*)row[1]);
												}
												else{
													average = 256;
												}

												if (instance.debug) {
													fprintf(stderr, "DEBUG: MYSQL result: %s\n",(char*)row[0]);
													fprintf(stderr, "DEBUG: MYSQL result: %d\n",average);
												}

												found_srv++;
												sprintf(key_srv_seq, "%s_%hu", key_srv, i);
												if ((strstr(input.qname, "._tls.") != NULL) || (strstr(input.qname, "_sips._tcp.") != NULL)){
													sprintf(ipstr, "%i\t1 5061 %s.", average, hostname_srv);
												}
												else{
													sprintf(ipstr, "%i\t1 5060 %s.", average, hostname_srv);
												}
												rc = memcached_set(memc, key_srv_seq, strlen(key_srv_seq), ipstr, strlen(ipstr),(time_t)instance.memcached.ttl, (uint32_t)0);

												if (instance.debug) {
													if (rc == MEMCACHED_SUCCESS) 
														fprintf(stderr, "DEBUG: memcached key: %s set with value '%s'\n", key_srv_seq, ipstr);
													else
														fprintf(stderr, "DEBUG: memcached key: %s was not able to be set into memcached with value '%s'\n", key_srv_seq, ipstr);
												}

												if (instance.debug) {
													fprintf(stderr, "DEBUG:DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
												}
												printf ("DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);

											}
											mysql_free_result((MYSQL_RES*)resource);
										}
										else{
											int rows;
											unsigned int i;
											unsigned int j;
											rows = PQntuples((PGresult*)resource);

											for (j = 0, i = 1; i < rows; j++, i++){
												strcpy(hostname_srv, PQgetvalue((PGresult*)resource,j,0));
												if (PQgetvalue((PGresult*)resource,j,1)){
													average = atoi(PQgetvalue((PGresult*)resource,j,1));
												}
												else{
													average = 256;
												}

												if (instance.debug) {
													fprintf(stderr, "DEBUG: PGSQL result: %s\n",(char*)PQgetvalue((PGresult*)resource,j,0));
													fprintf(stderr, "DEBUG: PGSQL result: %d\n",average);
												}
												found_srv++;

												sprintf(key_srv_seq, "%s_%hu", key_srv, i);
												if ((strstr(input.qname, "._tls.") != NULL) || (strstr(input.qname, "_sips._tcp.") != NULL)){
													sprintf(ipstr, "%i\t1 5061 %s.", average, hostname_srv);
												}
												else{
													sprintf(ipstr, "%i\t1 5060 %s.", average, hostname_srv);
												}
												rc = memcached_set(memc, key_srv_seq, strlen(key_srv_seq), ipstr, strlen(ipstr),(time_t)instance.memcached.ttl, (uint32_t)0);
												if (instance.debug) {
													if (rc == MEMCACHED_SUCCESS) 
														fprintf(stderr, "DEBUG: memcached key: %s set with value '%s'\n", key_srv, ipstr);
													else
														fprintf(stderr, "DEBUG: memcached key: %s was not able to be set into memcached with value '%s'\n", key_srv, ipstr);
												}

												if (instance.debug) {
													fprintf(stderr, "DEBUG:DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
												}
												printf ("DATA\t%s\tin\tSRV\t%i\t1\t%s\n", input.qname, instance.dns.srv_ttl, (char *)ipstr);
											}
											PQclear((PGresult*)resource);
										}

										free(hostname_srv);
										free(ipstr);
									}
								}
								free(key_srv_type);
								free(key_srv_seq);
								free(key_srv);
							}

							if ((instance.numeric_version >= 10200) &&
								((strncmp(input.qtype, "any", 3) == 0) || (strncmp(input.qtype, "naptr", 5) == 0))
							){
								if (instance.debug) {
									fprintf(stderr, "DATA\t%s\t%s\tNAPTR\t600\t-1\t10 100 \"S\" \"SIP+D2U\" \"\" _sip._udp.%s.\n", input.qname, input.qclass, input.qname);
									fprintf(stderr, "DATA\t%s\t%s\tNAPTR\t600\t-1\t10 100 \"S\" \"SIP+D2T\" \"\" _sip._tcp.%s.\n", input.qname, input.qclass, input.qname);
								}
								printf ("DATA\t%s\t%s\tNAPTR\t600\t-1\t10 100 \"S\" \"SIP+D2U\" \"\" _sip._udp.%s.\n", input.qname, input.qclass, input.qname);
								printf ("DATA\t%s\t%s\tNAPTR\t600\t-1\t10 100 \"S\" \"SIP+D2T\" \"\" _sip._tcp.%s.\n", input.qname, input.qclass, input.qname);
							}
								
							if (
//								(found_a|found_srv) && (
											(strncmp(input.qtype, "any", 3) == 0) || 
											(strncmp(input.qtype, "soa", 3) == 0) // )
							) {
								if (instance.debug) {
									fprintf(stderr, "DEBUG:DATA\t%s\t%s\tSOA\t3600\t-1\t%s. %s. 2013112701 1800 3600 604800 3600\n", input.qname, input.qclass, instance.dns.email, instance.dns.ns1);
								}
								printf ("DATA\t%s\t%s\tSOA\t3600\t-1\t%s. %s. 2013112701 1800 3600 604800 3600\n", input.qname, input.qclass, instance.dns.email, instance.dns.ns1);
							}

							if (
								(found_a|found_srv) && (
									(strncmp(input.qtype, "any", 3) == 0) || 
									(strncmp(input.qtype, "ns", 2) == 0)
								)
								) {
								if (instance.debug) {
									fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns1);
									if (strlen(instance.dns.ns2))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns2);
									if (strlen(instance.dns.ns3))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns3);
									if (strlen(instance.dns.ns4))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns4);
									if (strlen(instance.dns.ns5))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns5);
									if (strlen(instance.dns.ns6))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns6);
									if (strlen(instance.dns.ns7))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns7);
									if (strlen(instance.dns.ns8))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns8);
									if (strlen(instance.dns.ns9))
										fprintf(stderr, "DEBUG:DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns9);
								}
								printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns1);
								if (strlen(instance.dns.ns2))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns2);
								if (strlen(instance.dns.ns3))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns3);
								if (strlen(instance.dns.ns4))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns4);
								if (strlen(instance.dns.ns5))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns5);
								if (strlen(instance.dns.ns6))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns6);
								if (strlen(instance.dns.ns7))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns7);
								if (strlen(instance.dns.ns8))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns8);
								if (strlen(instance.dns.ns9))
									printf ("DATA\t%s\t%s\tNS\t60\t-1\t%s\n", input.qname, input.qclass, instance.dns.ns9);
							}
						} // if (ipcount)
					}
				} // if (connected)
				else{
					if (instance.debug) {
						fprintf(stderr, "DEBUG:No Database...\n");
					}
				}
			} // if (strstr(input.qname, instance.dns.zone) != NULL)
			else{
				if (instance.debug) {
					fprintf(stderr, "DEBUG: Couldnt find %s in %s\n", instance.dns.zone, input.qname);
				}
			}

			if (instance.debug) {
				fprintf(stderr, "DEBUG:END\n");
			}
			printf ("END\n");
		}
	}
	
	if (instance.database.type==t_MYSQL){
		mysql_close((MYSQL*)conn);
	}
	else{
		PQfinish((PGconn*)conn);
	}
	fflush(stdin);
	fflush(stdout);
	return 0;
}
